/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include <cstring>
#include <fstream>
#include <iostream>

#include "Exception.h"
#include "Aircraft.h"
#include "CRC32.h"
#include "Util.h"

const char Aircraft::AirspeedUnitsText[][10]={"KPH","MPH","Knots"};
const char Aircraft::ClimbDescentRateUnitsText[][20]={"Meters/Sec","Feet/Min","Knots"};
const char Aircraft::AltitudeUnitsText[][10]={"Meters","Feet"};
const char Aircraft::AircraftWeightUnitsText[][10]={"Kilograms","Pounds"};
const char Aircraft::CGStationDistUnitsText[][20]={"Meters","Centimeters","Feet","Inches"};
const char Aircraft::OilQtyUnitsText[][10]={"Kilograms","Liters","Pounds","Gallons","Quarts"};
const char Aircraft::FuelQtyUnitsText[][10]={"Kilograms","Liters","Pounds","Gallons"};
const char Aircraft::FuelTypeText[][10]={"AvGas","MoGas","Jet"};
const double Aircraft::FuelTypeDensity[] = {0.7214,0.7214,0.8075};	/* kg/L one for each FuelTypeEnum */

Aircraft::Aircraft()
{
	Desc="New Aircraft";
	Registration="N0000";
	Filename="";

	TASClimb=KT2KPH(80.0);
	TASCruise=KT2KPH(120.0);
	TASDescent=KT2KPH(135.0);
	ClimbRate=FTPERMIN2MPERS(500.0);
	DescentRate=FTPERMIN2MPERS(500.0);
	FuelRateModel=FUELRATEMODEL_SEGMENT;
	memset(&FuelRate,0,sizeof(FuelRate));
	FuelRate.Segment.Climb=LAVGAS2KG(GAL2L(12.0));
	FuelRate.Segment.Cruise=LAVGAS2KG(GAL2L(10.0));
	FuelRate.Segment.Descent=LAVGAS2KG(GAL2L(8.0));

	AirspeedUnits=AIRSPEEDUNITS_KNOTS;
	ClimbDescentRateUnits=CLIMBDESCENTRATEUNITS_FTPERMIN;
	AltitudeUnits=ALTITUDEUNITS_FEET;
	AircraftWeightUnits=AIRCRAFTWEIGHTUNITS_LB;
	CGStationDistUnits=CGSTATIONDISTUNITS_IN;
	OilQtyUnits=OILQTYUNITS_QT;
	FuelQtyUnits=FUELQTYUNITS_GAL;
	FuelType=FUELTYPE_AVGAS;
}

Aircraft::~Aircraft()
{
	for (std::vector<WBEnvelope*>::iterator i=WBEnvelopes.begin();i!=WBEnvelopes.end();i++) {
		delete *i;
	}
	WBEnvelopes.clear();
}

std::string Aircraft::GetDesc() {
	return Desc;
}

void Aircraft::SetDesc(std::string value) {
	Desc=value;
}

std::string Aircraft::GetRegistration() {
	return Registration;
}

void Aircraft::SetRegistration(std::string value) {
	Registration=value;
}

double Aircraft::GetTASClimb() {
	return TASClimb;
}

void Aircraft::SetTASClimb(double value) {
	TASClimb=value;
}

double Aircraft::GetTASCruise() {
	return TASCruise;
}

void Aircraft::SetTASCruise(double value) {
	TASCruise=value;
}

double Aircraft::GetTASDescent() {
	return TASDescent;
}

void Aircraft::SetTASDescent(double value) {
	TASDescent=value;
}

double Aircraft::GetClimbRate() {
	return ClimbRate;
}

void Aircraft::SetClimbRate(double value) {
	ClimbRate=value;
}

double Aircraft::GetDescentRate() {
	return DescentRate;
}

void Aircraft::SetDescentRate(double value) {
	DescentRate=value;
}

enum Aircraft::FuelRateModelEnum Aircraft::GetFuelRateModel() {
	return FuelRateModel;
}

void Aircraft::SetFuelRateModel(enum FuelRateModelEnum value) {
	if (value<FUELRATEMODEL_INVALID)
		FuelRateModel=value;
}

double Aircraft::GetFuelRateClimb() {
	if (FuelRateModel==FUELRATEMODEL_SEGMENT)
		return FuelRate.Segment.Climb;
	else
		return 0.0;
}

void Aircraft::SetFuelRateClimb(double value) {
	SetFuelRateModel(FUELRATEMODEL_SEGMENT);
	FuelRate.Segment.Climb=value;
}

double Aircraft::GetFuelRateCruise() {
	if (FuelRateModel==FUELRATEMODEL_SEGMENT)
		return FuelRate.Segment.Cruise;
	else
		return 0.0;
}

void Aircraft::SetFuelRateCruise(double value) {
	SetFuelRateModel(FUELRATEMODEL_SEGMENT);
	FuelRate.Segment.Cruise=value;
}

double Aircraft::GetFuelRateDescent() {
	if (FuelRateModel==FUELRATEMODEL_SEGMENT)
		return FuelRate.Segment.Descent;
	else
		return 0.0;
}

void Aircraft::SetFuelRateDescent(double value) {
	SetFuelRateModel(FUELRATEMODEL_SEGMENT);
	FuelRate.Segment.Descent=value;
}

double Aircraft::GetFuelRateHourly(unsigned int Hour) {
	if (Hour>=FUELRATEMODEL_HOURLY_NHOURS)
		return 0.0;

	if (FuelRateModel==FUELRATEMODEL_HOURLY)
		return FuelRate.Hour[Hour];
	else
		return 0.0;
}

void Aircraft::SetFuelRateHourly(double value, unsigned int Hour) {
	if (Hour<FUELRATEMODEL_HOURLY_NHOURS) {
		SetFuelRateModel(FUELRATEMODEL_HOURLY);
		FuelRate.Hour[Hour]=value;
	}
}

enum Aircraft::AirspeedUnitsEnum Aircraft::GetAirspeedUnits() {
	return AirspeedUnits;
}

void Aircraft::SetAirspeedUnits(enum AirspeedUnitsEnum value) {
	if (value<AIRSPEEDUNITS_INVALID)
		AirspeedUnits=value;
}

enum Aircraft::ClimbDescentRateUnitsEnum Aircraft::GetClimbDescentRateUnits() {
	return ClimbDescentRateUnits;
}

void Aircraft::SetClimbDescentRateUnits(enum ClimbDescentRateUnitsEnum value) {
	if (value<CLIMBDESCENTRATEUNITS_INVALID)
		ClimbDescentRateUnits=value;
}

enum Aircraft::AltitudeUnitsEnum Aircraft::GetAltitudeUnits() {
	return AltitudeUnits;
}

void Aircraft::SetAltitudeUnits(enum AltitudeUnitsEnum value) {
	if (value<ALTITUDEUNITS_INVALID)
		AltitudeUnits=value;
}

enum Aircraft::AircraftWeightUnitsEnum Aircraft::GetAircraftWeightUnits() {
	return AircraftWeightUnits;
}

void Aircraft::SetAircraftWeightUnits(enum AircraftWeightUnitsEnum value) {
	if (value<AIRCRAFTWEIGHTUNITS_INVALID)
		AircraftWeightUnits=value;
}

enum Aircraft::CGStationDistUnitsEnum Aircraft::GetCGStationDistUnits() {
	return CGStationDistUnits;
}

void Aircraft::SetCGStationDistUnits(enum CGStationDistUnitsEnum value) {
	if (value<CGSTATIONDISTUNITS_INVALID)
		CGStationDistUnits=value;
}

enum Aircraft::OilQtyUnitsEnum Aircraft::GetOilQtyUnits() {
	return OilQtyUnits;
}

void Aircraft::SetOilQtyUnits(enum OilQtyUnitsEnum value) {
	if (value<OILQTYUNITS_INVALID)
		OilQtyUnits=value;
}

enum Aircraft::FuelQtyUnitsEnum Aircraft::GetFuelQtyUnits() {
	return FuelQtyUnits;
}

void Aircraft::SetFuelQtyUnits(enum FuelQtyUnitsEnum value) {
	if (value<FUELQTYUNITS_INVALID)
		FuelQtyUnits=value;
}

enum Aircraft::FuelTypeEnum Aircraft::GetFuelType() {
	return FuelType;
}

void Aircraft::SetFuelType(enum FuelTypeEnum value) {
	if (value<FUELTYPE_INVALID)
		FuelType=value;
}

std::string Aircraft::GetFilename() {
	return Filename;
}

bool Aircraft::Load(std::string Filename) {
	std::ifstream ifs;
	struct Aircraft_Binary_struct BinaryData;
	CRC32 CRC;
	uint32_t StoredCRC32;

	this->Filename=Filename;

	memset(&BinaryData,0,sizeof(BinaryData));

	ifs.open(this->Filename.c_str(),std::ifstream::binary);
	if (!ifs.good()) {
		throw Exception("Error opening "+this->Filename+": "+strerror(errno),Exception::LEVEL_ERROR);
	}

	/* Load the data */
	try {
		ifs.read((char*)&BinaryData,sizeof(BinaryData));
		if (!ifs.good()) {
			throw Exception("Error reading aircraft data",Exception::LEVEL_ERROR);
		}

		if (strncmp(BinaryData.Magic,AIRCRAFT_FILEMAGIC,sizeof(AIRCRAFT_FILEMAGIC)!=0)) {
			throw Exception("File corrupt: bad magic in "+Filename,Exception::LEVEL_ERROR);
		}
		if (BinaryData.DataVersion!=AIRCRAFT_BINARY_STRUCT_VER) {
			throw Exception("Unsupported data version in "+Filename,Exception::LEVEL_ERROR);
		}
		/* Check the CRC */
		StoredCRC32=BinaryData.CRC32;
		BinaryData.CRC32=0;

		if (StoredCRC32!=CRC.Finalize(CRC.Compute(&BinaryData,sizeof(BinaryData)))) {
			throw Exception("File corrupt - invalid CRC in "+this->Filename,Exception::LEVEL_ERROR);
		}

		SetFromBinary(&BinaryData);
	} catch (const Exception& Ex) {
		ifs.close();
		throw Ex;
	}

	ifs.close();
	return true;
}

bool Aircraft::Load() {
	/* Stub */
	return Load(Filename);
}

void Aircraft::Save(std::string Filename) {
	std::ofstream ofs;
	struct Aircraft_Binary_struct BinaryData;
	CRC32 CRC;

	this->Filename=Filename;

	if (Filename.compare("")!=0) {
		Util::ShellExpand(&Filename);
		this->Filename=Filename;
	} else {
		Filename=this->Filename;
	}

	if (!MakeBinary(&BinaryData)) {
		throw Exception("Error creating "+Filename+": "+strerror(errno),Exception::LEVEL_ERROR);
	}

	ofs.open(Filename.c_str(),std::ofstream::binary);
	if (!ofs.good()) {
		throw Exception("Error opening "+Filename+": "+strerror(errno),Exception::LEVEL_ERROR);
	}

	if (!ofs.good()) {
		ofs.close();
		throw Exception("Error processing "+Filename,Exception::LEVEL_ERROR);
	}

	BinaryData.CRC32=CRC.Finalize(CRC.Compute(&BinaryData,sizeof(BinaryData)));
	ofs.write((char*)&BinaryData,sizeof(BinaryData));
	if (!ofs.good()) {
		ofs.close();
		throw Exception("Error writing "+Filename,Exception::LEVEL_ERROR);
	}

	ofs.close();
}

void Aircraft::Save() {
	Save(Filename);
}

bool Aircraft::MakeBinary(struct Aircraft_Binary_struct *BinaryData) {
	unsigned int i;

	/* Sanity checking */
	if (
		Desc.size()>sizeof(BinaryData->Desc) ||
		Registration.size()>sizeof(BinaryData->Registration) ||
		TASClimb*10000>UINT32_MAX ||
		TASCruise*10000>UINT32_MAX ||
		TASDescent*10000>UINT32_MAX ||
		ClimbRate*10000>UINT32_MAX ||
		DescentRate*10000>UINT32_MAX ||
		FuelRateModel>UINT8_MAX ||
		FuelRateModel>=FUELRATEMODEL_INVALID ||
		FuelRate.Segment.Climb*10000>UINT32_MAX ||
		FuelRate.Segment.Cruise*10000>UINT32_MAX ||
		FuelRate.Segment.Descent*10000>UINT32_MAX ||
		/* FuelRateHour is checked below since it's an array */
		AirspeedUnits>UINT8_MAX ||
		AirspeedUnits>=AIRSPEEDUNITS_INVALID ||
		ClimbDescentRateUnits>UINT8_MAX ||
		ClimbDescentRateUnits>=CLIMBDESCENTRATEUNITS_INVALID ||
		AltitudeUnits>UINT8_MAX ||
		AltitudeUnits>=ALTITUDEUNITS_INVALID ||
		AircraftWeightUnits>UINT8_MAX ||
		AircraftWeightUnits>=AIRCRAFTWEIGHTUNITS_INVALID ||
		CGStationDistUnits>UINT8_MAX ||
		CGStationDistUnits>=CGSTATIONDISTUNITS_INVALID ||
		OilQtyUnits>UINT8_MAX ||
		OilQtyUnits>=OILQTYUNITS_INVALID ||
		FuelQtyUnits>UINT8_MAX ||
		FuelQtyUnits>=FUELQTYUNITS_INVALID ||
		FuelType>UINT8_MAX ||
		FuelType>=FUELTYPE_INVALID
		) {
		return false;
	}
	for (i=0;i<FUELRATEMODEL_HOURLY_NHOURS;i++) {
		if (FuelRate.Hour[i]*10000>UINT32_MAX)
			return false;
	}

	memset(BinaryData,0,sizeof(*BinaryData));
	strncpy(BinaryData->Magic,AIRCRAFT_FILEMAGIC,sizeof(AIRCRAFT_FILEMAGIC));
	BinaryData->DataVersion=AIRCRAFT_BINARY_STRUCT_VER;
	strncpy(BinaryData->Desc,Desc.c_str(),sizeof(BinaryData->Desc));
	strncpy(BinaryData->Registration,Registration.c_str(),sizeof(BinaryData->Registration));
	BinaryData->TASClimbx10000=TASClimb*10000;
	BinaryData->TASCruisex10000=TASCruise*10000;
	BinaryData->TASDescentx10000=TASDescent*10000;
	BinaryData->ClimbRatex10000=ClimbRate*10000;
	BinaryData->DescentRatex10000=DescentRate*10000;
	BinaryData->FuelRateModel=FuelRateModel;
	BinaryData->FuelRateClimbx10000=FuelRate.Segment.Climb*10000;
	BinaryData->FuelRateCruisex10000=FuelRate.Segment.Cruise*10000;
	BinaryData->FuelRateDescentx10000=FuelRate.Segment.Descent*10000;
	for (i=0;i<FUELRATEMODEL_HOURLY_NHOURS;i++) {
		BinaryData->FuelRateHourx10000[i]=FuelRate.Hour[i]*10000;
	}
	BinaryData->AirspeedUnits=AirspeedUnits;
	BinaryData->ClimbDescentRateUnits=ClimbDescentRateUnits;
	BinaryData->AltitudeUnits=AltitudeUnits;
	BinaryData->AircraftWeightUnits=AircraftWeightUnits;
	BinaryData->CGStationDistUnits=CGStationDistUnits;
	BinaryData->OilQtyUnits=OilQtyUnits;
	BinaryData->FuelQtyUnits=FuelQtyUnits;
	BinaryData->FuelType=FuelType;

	return true;
}

void Aircraft::SetFromBinary(struct Aircraft_Binary_struct *BinaryData) {
	unsigned int i;

	Desc.assign(BinaryData->Desc,sizeof(BinaryData->Desc));
	Registration.assign(BinaryData->Registration,sizeof(BinaryData->Registration));
	TASClimb=BinaryData->TASClimbx10000/10000.0;
	TASCruise=BinaryData->TASCruisex10000/10000.0;
	TASDescent=BinaryData->TASDescentx10000/10000.0;
	ClimbRate=BinaryData->ClimbRatex10000/10000.0;
	DescentRate=BinaryData->DescentRatex10000/10000.0;
	FuelRateModel=(Aircraft::FuelRateModelEnum)BinaryData->FuelRateModel;
	FuelRate.Segment.Climb=BinaryData->FuelRateClimbx10000/10000.0;
	FuelRate.Segment.Cruise=BinaryData->FuelRateCruisex10000/10000.0;
	FuelRate.Segment.Descent=BinaryData->FuelRateDescentx10000/10000.0;
	for (i=0;i<FUELRATEMODEL_HOURLY_NHOURS;i++) {
		FuelRate.Hour[i]=BinaryData->FuelRateHourx10000[i]/10000.0;
	}
	AirspeedUnits=(Aircraft::AirspeedUnitsEnum)BinaryData->AirspeedUnits;
	ClimbDescentRateUnits=(Aircraft::ClimbDescentRateUnitsEnum)BinaryData->ClimbDescentRateUnits;
	AltitudeUnits=(Aircraft::AltitudeUnitsEnum)BinaryData->AltitudeUnits;
	AircraftWeightUnits=(Aircraft::AircraftWeightUnitsEnum)BinaryData->AircraftWeightUnits;
	CGStationDistUnits=(Aircraft::CGStationDistUnitsEnum)BinaryData->CGStationDistUnits;
	OilQtyUnits=(Aircraft::OilQtyUnitsEnum)BinaryData->OilQtyUnits;
	FuelQtyUnits=(Aircraft::FuelQtyUnitsEnum)BinaryData->FuelQtyUnits;
	FuelType=(Aircraft::FuelTypeEnum)BinaryData->FuelType;

	/* Reset the size to the actual number of characters */
	Desc=Desc.c_str();
	Registration=Registration.c_str();
}

WBStation::WBStation() {
	Mass=0.0;
	Arm=0.0;
	Type=STATIONTYPE_INVALID;
}

void WBStation::SetByArm(enum StationTypeEnum Type, double Mass, double Arm) {
	this->Type=Type;
	this->Mass=Mass;
	this->Arm=Arm;
}

void WBStation::SetByMom(enum StationTypeEnum Type, double Mass, double Moment) {
	SetByArm(Type,Mass,Moment/Mass);
}

double WBStation::GetMass() {
	if (Type>=STATIONTYPE_INVALID)
		throw Exception("Weight and Balance station invalid",Exception::LEVEL_ERROR);

	return Mass;
}

double WBStation::GetArm() {
	if (Type>=STATIONTYPE_INVALID)
		throw Exception("Weight and Balance station invalid",Exception::LEVEL_ERROR);

	return Arm;
}

double WBStation::GetMoment() {
	if (Type>=STATIONTYPE_INVALID)
		throw Exception("Weight and Balance station invalid",Exception::LEVEL_ERROR);

	return Mass*Arm;
}
